// This unit defines all blocktypes
// As I wasn't able to develop a generic algorithm
// that would work for the rotation of all block types, I decided
// to hardcode the various rotation states.
//
// It seems that even the "original" Gameboy Tetris did this.
//
// If you don't understand the "naming scheme" of the blocks
//  -> http://de.wikipedia.org/w/index.php?title=Datei:Tetrominoes_IJLO_STZ_Worlds.svg

unit Blocks;

interface

uses
	Types, // Common FreePaTris Types
	SysUtils, // Exceptions
	Crt; // Console operations
	
// Creates a new block
function CreateBlock(id : TBlockId) : TBlock;

// Returns a random block ID
function CreateRandomBlockId() : TBlockId;

// Rotates the given block
procedure RotateBlock(clockwise : boolean; var block : TBlock);

implementation

// Calculates the boundary of the given block
//
// Parameters
//  block: The block that needs to have boundaries recalculated
//         The result is written into the block
procedure CalculateBoundary(var _block : TBlock);
var
	// The calculated boundaries
	_boundaryUpper, _boundaryLeft, _boundaryBottom, _boundaryRight : byte;
	_x, _y : byte; // Counting variables
begin
	// Assume maximum width  and height for start
	_boundaryUpper := Length(_block.Cells[_block.Orientation, 0]) - 1;
	_boundaryLeft := Length(_block.Cells[_block.Orientation]) - 1;
	
	// Assume minimum bottom and right for start
	_boundaryBottom := 0;
	_boundaryRight := 0;

	// Iterate over block content
	for _x := 0 to Length(_block.Cells[_block.Orientation]) - 1 do 
	begin
		for _y := 0 to Length(_block.Cells[_block.Orientation, _x]) - 1 do 
		begin	
			// Check whether the block blocks at the block position
			if _block.Cells[_block.Orientation, _x, _y] then 
			begin
				// Is the cell more left then the current left boundary
				if (_x < _boundaryLeft) then 
					_boundaryLeft := _x;
				// Is the cell higher then the current upper boundary
				if (_y < _boundaryUpper) then 
					_boundaryUpper := _y;
				// Is the cell more right then the current right boundary
				if (_x > _boundaryRight) then
					_boundaryRight := _x;
				// Is the cell lower then the current lower boundary
				if (_y > _boundaryBottom) then
					_boundaryBottom := _y;
			end;
		end;
	end;

	// Copy the computed values into the block
	_block.Boundary.Upper := _boundaryUpper;
	_block.Boundary.Left := _boundaryLeft;
	_block.Boundary.Lower := _boundaryBottom;
	_block.Boundary.Right := _boundaryRight;
end;

// Rotates the given block either clockwise or anticlockwise
//
// Parameters
//  clockwise: Specifies the rotation direction
//  block: The block that shal be rotated
//
// Calls
//  - CalculateBoundary
procedure RotateBlock(clockwise : boolean; var block : TBlock);
begin
	// Turn clockwise
	if clockwise then
	begin
        // Adding 1 to the orientation enum turns clockwise
		block.Orientation := block.Orientation + 1;
		if (block.Orientation > byte(Left)) then
		begin
			// Reset to Up if turned to far
			block.Orientation := byte(Up);
		end
    // Turn anticlockwise
	end else
	begin
        // Subtracting 1 from the orientation enum turns counter clockwise
		block.Orientation := block.Orientation - 1;
		if (block.Orientation < 0) then
		begin
			// Reset to Left if turned to far
			block.Orientation := byte(Left);
		end;
	end;
	
	// Re-calculate boundary
	CalculateBoundary(block);
end;

// Fill the given block cells with the
// 'O' values
//
// Parameters:
//  block: The block which cells shal be altered
procedure FillBlockO(var block : TBlock);
begin
	// Assigning the fields
	SetLength(block.Cells, 4, 2, 2);

	// ##
	// ##
	block.Cells[Byte(Up), 0,0] := true;
	block.Cells[Byte(Up), 0,1] := true;
	block.Cells[Byte(Up), 1,0] := true;
	block.Cells[Byte(Up), 1,1] := true;

	// Other rotations can be copied
	block.Cells[byte(Right)] := block.Cells[Byte(Up)];
	block.Cells[byte(Down)] := block.Cells[Byte(Up)];
	block.Cells[byte(Left)] := block.Cells[Byte(Up)];
end;

// Fill the given block cells with the
// 'I' values
//
// Parameters:
//  block: The block which cells shal be altered
procedure FillBlockI(var block : TBlock);
begin
	SetLength(block.Cells, 4, 4, 4);

	// .#..
	// .#..
	// .#..
	// .#..
	block.Cells[Byte(Up), 0, 0] := false;
	block.Cells[Byte(Up), 1, 0] := true;
	block.Cells[Byte(Up), 2, 0] := false;
	block.Cells[Byte(Up), 3, 0] := false;
	block.Cells[Byte(Up), 0, 1] := false;
	block.Cells[Byte(Up), 1, 1] := true;
	block.Cells[Byte(Up), 2, 1] := false;
	block.Cells[Byte(Up), 3, 1] := false;
	block.Cells[Byte(Up), 0, 2] := false;
	block.Cells[Byte(Up), 1, 2] := true;
	block.Cells[Byte(Up), 2, 2] := false;
	block.Cells[Byte(Up), 3, 2] := false;
	block.Cells[Byte(Up), 0, 3] := false;
	block.Cells[Byte(Up), 1, 3] := true;
	block.Cells[Byte(Up), 2, 3] := false;
	block.Cells[Byte(Up), 3, 3] := false;

	// ....
	// ####
	// ....
	// ....
	block.Cells[Byte(Right), 0, 0] := false;
	block.Cells[Byte(Right), 1, 0] := false;
	block.Cells[Byte(Right), 2, 0] := false;
	block.Cells[Byte(Right), 3, 0] := false;
	block.Cells[Byte(Right), 0, 1] := true;
	block.Cells[Byte(Right), 1, 1] := true;
	block.Cells[Byte(Right), 2, 1] := true;
	block.Cells[Byte(Right), 3, 1] := true;
	block.Cells[Byte(Right), 0, 2] := false;
	block.Cells[Byte(Right), 1, 2] := false;
	block.Cells[Byte(Right), 2, 2] := false;
	block.Cells[Byte(Right), 3, 2] := false;
	block.Cells[Byte(Right), 0, 3] := false;
	block.Cells[Byte(Right), 1, 3] := false;
	block.Cells[Byte(Right), 2, 3] := false;
	block.Cells[Byte(Right), 3, 3] := false;

	// Other rotations can be copied
	block.Cells[Byte(Down)] := block.Cells[Byte(Up)];
    block.Cells[Byte(Left)] := block.Cells[Byte(Right)];
end;

// Fill the given block cells with the
// 'T' values
//
// Parameters:
//  block: The block which cells shal be altered
procedure FillBlockT(var block : TBlock);
begin
	SetLength(block.Cells, 4, 3, 3);

	// .#.
	// ###
	// ...
	block.Cells[Byte(Up), 0,0] := false;
	block.Cells[Byte(Up), 1,0] := true;
	block.Cells[Byte(Up), 2,0] := false;
	block.Cells[Byte(Up), 0,1] := true;
	block.Cells[Byte(Up), 1,1] := true;
	block.Cells[Byte(Up), 2,1] := true;
	block.Cells[Byte(Up), 0,2] := false;
	block.Cells[Byte(Up), 1,2] := false;
	block.Cells[Byte(Up), 2,2] := false;

	// .#.
	// .##
	// .#.
	block.Cells[Byte(Right), 0,0] := false;
	block.Cells[Byte(Right), 1,0] := true;
	block.Cells[Byte(Right), 2,0] := false;
	block.Cells[Byte(Right), 0,1] := false;
	block.Cells[Byte(Right), 1,1] := true;
	block.Cells[Byte(Right), 2,1] := true;
	block.Cells[Byte(Right), 0,2] := false;
	block.Cells[Byte(Right), 1,2] := true;
	block.Cells[Byte(Right), 2,2] := false;

	// ...
	// ###
	// .#.
	block.Cells[Byte(Down), 0,0] := false;
	block.Cells[Byte(Down), 1,0] := false;
	block.Cells[Byte(Down), 2,0] := false;
	block.Cells[Byte(Down), 0,1] := true;
	block.Cells[Byte(Down), 1,1] := true;
	block.Cells[Byte(Down), 2,1] := true;
	block.Cells[Byte(Down), 0,2] := false;
	block.Cells[Byte(Down), 1,2] := true;
	block.Cells[Byte(Down), 2,2] := false;

	// .#.
	// ##.
	// .#.
	block.Cells[Byte(Left), 0,0] := false;
	block.Cells[Byte(Left), 1,0] := true;
	block.Cells[Byte(Left), 2,0] := false;
	block.Cells[Byte(Left), 0,1] := true;
	block.Cells[Byte(Left), 1,1] := true;
	block.Cells[Byte(Left), 2,1] := false;
	block.Cells[Byte(Left), 0,2] := false;
	block.Cells[Byte(Left), 1,2] := true;
	block.Cells[Byte(Left), 2,2] := false;
end;

// Fill the given block cells with the
// 'S' values
//
// Parameters:
//  block: The block which cells shal be altered
procedure FillBlockS(var block : TBlock);
begin
	SetLength(block.Cells, 4, 3, 3);

	// .##
	// ##.
	// ...
	block.Cells[Byte(Up), 0,0] := false;
	block.Cells[Byte(Up), 1,0] := true;
	block.Cells[Byte(Up), 2,0] := true;
	block.Cells[Byte(Up), 0,1] := true;
	block.Cells[Byte(Up), 1,1] := true;
	block.Cells[Byte(Up), 2,1] := false;
	block.Cells[Byte(Up), 0,2] := false;
	block.Cells[Byte(Up), 1,2] := false;
	block.Cells[Byte(Up), 2,2] := false;

	// .#.
	// .##
	// ..#
	block.Cells[Byte(Right), 0,0] := false;
	block.Cells[Byte(Right), 1,0] := true;
	block.Cells[Byte(Right), 2,0] := false;
	block.Cells[Byte(Right), 0,1] := false;
	block.Cells[Byte(Right), 1,1] := true;
	block.Cells[Byte(Right), 2,1] := true;
	block.Cells[Byte(Right), 0,2] := false;
	block.Cells[Byte(Right), 1,2] := false;
	block.Cells[Byte(Right), 2,2] := true;

    // Other rotations can be copied
	block.Cells[Byte(Down)] := block.Cells[Byte(Up)];
    block.Cells[Byte(Left)] := block.Cells[Byte(Right)];
end;

// Fill the given block cells with the
// 'Z' values
//
// Parameters:
//  block: The block which cells shal be altered
procedure FillBlockZ(var block : TBlock);
begin
	SetLength(block.Cells, 4, 3, 3);

	// ##.
	// .##
	// ...
	block.Cells[Byte(Up), 0,0] := true;
	block.Cells[Byte(Up), 1,0] := true;
	block.Cells[Byte(Up), 2,0] := false;
	block.Cells[Byte(Up), 0,1] := false;
	block.Cells[Byte(Up), 1,1] := true;
	block.Cells[Byte(Up), 2,1] := true;
	block.Cells[Byte(Up), 0,2] := false;
	block.Cells[Byte(Up), 1,2] := false;
	block.Cells[Byte(Up), 2,2] := false;

	// .#.
	// ##.
	// #..
	block.Cells[Byte(Right), 0,0] := false;
	block.Cells[Byte(Right), 1,0] := true;
	block.Cells[Byte(Right), 2,0] := false;
	block.Cells[Byte(Right), 0,1] := true;
	block.Cells[Byte(Right), 1,1] := true;
	block.Cells[Byte(Right), 2,1] := false;
	block.Cells[Byte(Right), 0,2] := true;
	block.Cells[Byte(Right), 1,2] := false;
	block.Cells[Byte(Right), 2,2] := false;

    // Other rotations can be copied
	block.Cells[Byte(Down)] := block.Cells[Byte(Up)];
    block.Cells[Byte(Left)] := block.Cells[Byte(Right)];
end;

// Fill the given block cells with the
// 'J' values
//
// Parameters:
//  block: The block which cells shal be altered
procedure FillBlockJ(var block : TBlock);
begin
	SetLength(block.Cells, 4, 3, 3);

	// .#.
	// .#.
	// ##.
	block.Cells[Byte(Up), 0,0] := false;
	block.Cells[Byte(Up), 1,0] := true;
	block.Cells[Byte(Up), 2,0] := false;
	block.Cells[Byte(Up), 0,1] := false;
	block.Cells[Byte(Up), 1,1] := true;
	block.Cells[Byte(Up), 2,1] := false;
	block.Cells[Byte(Up), 0,2] := true;
	block.Cells[Byte(Up), 1,2] := true;
	block.Cells[Byte(Up), 2,2] := false;

	// #..
	// ###
	// ...
	block.Cells[Byte(Right), 0,0] := true;
	block.Cells[Byte(Right), 1,0] := false;
	block.Cells[Byte(Right), 2,0] := false;
	block.Cells[Byte(Right), 0,1] := true;
	block.Cells[Byte(Right), 1,1] := true;
	block.Cells[Byte(Right), 2,1] := true;
	block.Cells[Byte(Right), 0,2] := false;
	block.Cells[Byte(Right), 1,2] := false;
	block.Cells[Byte(Right), 2,2] := false;

	// .##
	// .#.
	// .#.
	block.Cells[Byte(Down), 0,0] := false;
	block.Cells[Byte(Down), 1,0] := true;
	block.Cells[Byte(Down), 2,0] := true;
	block.Cells[Byte(Down), 0,1] := false;
	block.Cells[Byte(Down), 1,1] := true;
	block.Cells[Byte(Down), 2,1] := false;
	block.Cells[Byte(Down), 0,2] := false;
	block.Cells[Byte(Down), 1,2] := true;
	block.Cells[Byte(Down), 2,2] := false;

	// ...
	// ###
	// ..#
	block.Cells[Byte(Left), 0,0] := false;
	block.Cells[Byte(Left), 1,0] := false;
	block.Cells[Byte(Left), 2,0] := false;
	block.Cells[Byte(Left), 0,1] := true;
	block.Cells[Byte(Left), 1,1] := true;
	block.Cells[Byte(Left), 2,1] := true;
	block.Cells[Byte(Left), 0,2] := false;
	block.Cells[Byte(Left), 1,2] := false;
	block.Cells[Byte(Left), 2,2] := true;
end;

// Fill the given block cells with the
// 'L' values
//
// Parameters:
//  block: The block which cells shal be altered
procedure FillBlockL(var block : TBlock);
begin
	SetLength(block.Cells, 4, 3, 3);

	// .#.
	// .#.
	// .##
	block.Cells[Byte(Up), 0,0] := false;
	block.Cells[Byte(Up), 1,0] := true;
	block.Cells[Byte(Up), 2,0] := false;
	block.Cells[Byte(Up), 0,1] := false;
	block.Cells[Byte(Up), 1,1] := true;
	block.Cells[Byte(Up), 2,1] := false;
	block.Cells[Byte(Up), 0,2] := false;
	block.Cells[Byte(Up), 1,2] := true;
	block.Cells[Byte(Up), 2,2] := true;

	// ...
	// ###
	// #..
	block.Cells[Byte(Right), 0,0] := false;
	block.Cells[Byte(Right), 1,0] := false;
	block.Cells[Byte(Right), 2,0] := false;
	block.Cells[Byte(Right), 0,1] := true;
	block.Cells[Byte(Right), 1,1] := true;
	block.Cells[Byte(Right), 2,1] := true;
	block.Cells[Byte(Right), 0,2] := true;
	block.Cells[Byte(Right), 1,2] := false;
	block.Cells[Byte(Right), 2,2] := false;

	// ##.
	// .#.
	// .#.
	block.Cells[Byte(Down), 0,0] := true;
	block.Cells[Byte(Down), 1,0] := true;
	block.Cells[Byte(Down), 2,0] := false;
	block.Cells[Byte(Down), 0,1] := false;
	block.Cells[Byte(Down), 1,1] := true;
	block.Cells[Byte(Down), 2,1] := false;
	block.Cells[Byte(Down), 0,2] := false;
	block.Cells[Byte(Down), 1,2] := true;
	block.Cells[Byte(Down), 2,2] := false;

	// ..#
	// ###
	// ...
	block.Cells[Byte(Left), 0,0] := false;
	block.Cells[Byte(Left), 1,0] := false;
	block.Cells[Byte(Left), 2,0] := true;
	block.Cells[Byte(Left), 0,1] := true;
	block.Cells[Byte(Left), 1,1] := true;
	block.Cells[Byte(Left), 2,1] := true;
	block.Cells[Byte(Left), 0,2] := false;
	block.Cells[Byte(Left), 1,2] := false;
	block.Cells[Byte(Left), 2,2] := false;
end;

// Returns a random block Id
function CreateRandomBlockId() : TBlockId;
begin
     case random(BlockCount - 1) of
          0: CreateRandomBlockId := TBlockId(char('I'));
          1: CreateRandomBlockId := TBlockId(char('J'));
          2: CreateRandomBlockId := TBlockId(char('L'));
          3: CreateRandomBlockId := TBlockId(char('O'));
          4: CreateRandomBlockId := TBlockId(char('S'));
          5: CreateRandomBlockId := TBlockId(char('T'));
          6: CreateRandomBlockId := TBlockId(char('Z'));
    end;
end;

// Returns a new block matching the given ID
//
// Paramters
//  id: The type of block that shal be created
//
// Calls
//  - The appropriate CreateBlock* Method
function CreateBlock(id : TBlockId) : TBlock;
var
	block : TBlock;
begin
	// Assign id
	block.Id := id;

	// Assign position
	block.X := (GamefieldWidth div 2) - 2; // "Approximate" middle for blocks
	block.Y := 0;

	// Assign initial orientation
	block.Orientation := Byte(DefaultBlockOrientation);
	
	// Filling
	case id of
		TBlockO: FillBlockO(block);
		TBlockT: FillBlockT(block);
		TBlockI: FillBlockI(block);
		TBlockS: FillBlockS(block);
		TBlockZ: FillBlockZ(block);
		TBlockJ: FillBlockJ(block);
		TBlockL: FillBlockL(block);
	else
		raise Exception.Create('Block not yet implemented');
	end;
	
	// Boundaries
	CalculateBoundary(block);
	
	// Assign colour
	repeat
        begin
		// +1 because no black blocks
		// Brown - 1 to get the "basic" colours
		block.Colour := Random(Brown - 1) + 1;
        end;
	until (block.Colour <> GamefieldBackground);

	// Return the block
	CreateBlock := block;
end;

end.
